{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "79bd76a6",
   "metadata": {},
   "source": [
    "# Golang Generics\n",
    "\n",
    "In the past, I have found it very frustrating to program in Go.\n",
    "To me, it feels like a language that heavily prioritizes quantity over quality.\n",
    "It is a bit safer and definitely more performant than Python, but nil pointers, a lack of proper ADTs, no tuples, ...\n",
    "I digress.\n",
    "\n",
    "However, recent (and wildly contentious) language changes have made it a lot less of a headache to use!\n",
    "As of Go 1.18, the language supports generics.\n",
    "This is an awesome feature, but it needs to be wielded carefully.\n",
    "\n",
    "When all you have is a hammer, everything is a nail.\n",
    "\n",
    "## What are generics\n",
    "\n",
    "I am not a type theorist, so take this with a lot of hand-waving.\n",
    "👋 👋 👋 \n",
    "\n",
    "If you have programmed in Python or Ruby, you should be quite familiar with generic types whether or not you realize it.\n",
    "Any time you write a function that works on a container, you are dealing with generics.\n",
    "Let's look at a Python example with type annotations:\n",
    "\n",
    "```python\n",
    "T = TypeVar('T')\n",
    "\n",
    "def count_items(objects: Iterable[T]) -> int:\n",
    "    count = 0\n",
    "    for object in objects:\n",
    "        count += 1\n",
    "    return count\n",
    "```\n",
    "\n",
    "I added type annotations to highlight what is happening conceptually.\n",
    "`count_items` doesn't care what is inside the container.\n",
    "It works at a higher level than the contents of a list.\n",
    "It only cares about the collection itself.\n",
    "The type parameter, `T`, is what is known as a **generic** type.\n",
    "\"Why does this matter?\", you may ask.\n",
    "Let's evaluate a common situation in Go.\n",
    "\n",
    "[Playground](https://play.golang.com/p/s-rtptDC-z5)\n",
    "\n",
    "```go\n",
    "package main\n",
    "\n",
    "import (\n",
    "\t\"encoding/json\"\n",
    "\t\"fmt\"\n",
    ")\n",
    "\n",
    "type Dog struct {\n",
    "\tId   int    `json:\"id\"`\n",
    "\tSays string `json:\"says\"`\n",
    "}\n",
    "\n",
    "type Cat struct {\n",
    "\tId              int    `json:\"id\"`\n",
    "\tSays            string `json:\"says\"`\n",
    "\tTimesMisbehaved int    `json:\"incidents\"`\n",
    "}\n",
    "\n",
    "func GetDog(id int) *Dog {\n",
    "\tfmt.Printf(\"Getting Dog %d...\\n\", id)\n",
    "\tnewDog := &Dog{}\n",
    "\t_ = json.Unmarshal([]byte(`{\"says\": \"woof\"}`), newDog)\n",
    "\tnewDog.Id = id\n",
    "\treturn newDog\n",
    "}\n",
    "\n",
    "func GetCat(id int) *Cat {\n",
    "\tfmt.Printf(\"Getting Cat %d...\\n\", id)\n",
    "\tnewCat := &Cat{}\n",
    "\t_ = json.Unmarshal([]byte(`{\"says\": \"meow\", \"incidents\": 12}`), newCat)\n",
    "\tnewCat.Id = id\n",
    "\treturn newCat\n",
    "}\n",
    "\n",
    "func main() {\n",
    "\tfmt.Printf(\"%v\\n\", *GetDog(42))\n",
    "\tfmt.Printf(\"%v\\n\", *GetCat(5))\n",
    "}\n",
    "```\n",
    "\n",
    "Wouldn't it be nice to have a single `GetById` function?\n",
    "Both of our \"getters\" take an ID, unmarshall some bytes, and return an object.\n",
    "Raymond Hettinger is beating a podium somewhere exclaiming, \"There must be a better way!\".\n",
    "Any proper Gopher will look at this and shout, \"Use an interface!\".\n",
    "Fair point, let's explore that.\n",
    "\n",
    "## Interfaces\n",
    "\n",
    "Go functions that leverage an interface usually have a signature along the lines of:\n",
    "\n",
    "```go\n",
    "type CanSayHello interface {\n",
    "    Hello() string\n",
    "}\n",
    "\n",
    "func SayHello(helloObj CanSayHello) string {\n",
    "    return helloObj.Hello()\n",
    "}\n",
    "```\n",
    "\n",
    "We are faced with a sort of chicken and egg problem.\n",
    "Ideally, we want a function that calls an endpoint and returns an object based on that endpoint.\n",
    "Something along the lines of `Get(\"cats/\", id) -> *Cat`.\n",
    "Keep in mind that you *cannot pass types* in Go as variables.\n",
    "I want to see if we can extract the object-specific functionality into its own function.\n",
    "To do this, we can use an interface, but the function must return some kind of interface as well.\n",
    "\n",
    "[Playground](https://play.golang.com/p/dz1eyI1J1Re)\n",
    "\n",
    "```go\n",
    "package main\n",
    "\n",
    "import (\n",
    "\t\"encoding/json\"\n",
    "\t\"fmt\"\n",
    ")\n",
    "\n",
    "type Marshallable interface {\n",
    "\tUnmarshal([]byte) Marshallable\n",
    "}\n",
    "\n",
    "type Dog struct {\n",
    "\tId   int    `json:\"id\"`\n",
    "\tSays string `json:\"says\"`\n",
    "}\n",
    "\n",
    "type Cat struct {\n",
    "\tId              int    `json:\"id\"`\n",
    "\tSays            string `json:\"says\"`\n",
    "\tTimesMisbehaved int    `json:\"incidents\"`\n",
    "}\n",
    "\n",
    "func (d *Dog) Unmarshal(input []byte) Marshallable {\n",
    "\t_ = json.Unmarshal(input, d)\n",
    "\treturn d\n",
    "}\n",
    "\n",
    "func (c *Cat) Unmarshal(input []byte) Marshallable {\n",
    "\t_ = json.Unmarshal(input, c)\n",
    "\treturn c\n",
    "}\n",
    "\n",
    "func GetAnimal(animal Marshallable, id int) Marshallable {\n",
    "\tfmt.Printf(\"Getting Animal ID %d...\\n\", id)\n",
    "\tswitch animal.(type) {\n",
    "\tcase *Cat:\n",
    "\t\tanimal.Unmarshal([]byte(`{\"says\": \"meow\", \"incidents\": 12}`))\n",
    "\tcase *Dog:\n",
    "\t\tanimal.Unmarshal([]byte(`{\"says\": \"woof\"}`))\n",
    "\t}\n",
    "\treturn animal\n",
    "}\n",
    "\n",
    "func main() {\n",
    "\tnewCat := &Cat{}\n",
    "\tfmt.Printf(\"%v\\n\", GetAnimal(newCat, 5))\n",
    "\t\n",
    "\tnewDog := &Dog{}\n",
    "\tfmt.Printf(\"%v\\n\", GetAnimal(newDog, 42))\n",
    "}\n",
    "```\n",
    "\n",
    "The maintenance burden of this version is arguably a little lower, but we still end up with a ton of boilerplate around the object handler.\n",
    "The above isn't a terribly satisfying solution, either.\n",
    "We could have used an empty interface as the return type for `Marshallable`, and we are stuck `switch`ing on the type in either case.\n",
    "Type assertions are evaluated at runtime, and this leaves us wondering if there is a safer way to perform our tricks.\n",
    "\n",
    "Can we do better?\n",
    "🧐\n",
    "That's the whole point of this post!\n",
    "Generics to the rescue!\n",
    "🥳🥳🥳\n",
    "\n",
    "## Generic version\n",
    "\n",
    "[Playground](https://play.golang.com/p/bm32Tj9Y8So)\n",
    "\n",
    "```go\n",
    "package main\n",
    "\n",
    "import (\n",
    "\t\"encoding/json\"\n",
    "\t\"fmt\"\n",
    ")\n",
    "\n",
    "type HasEndpoint interface {\n",
    "\tGetEndpoint() string\n",
    "}\n",
    "\n",
    "type Dog struct {\n",
    "\tId   int    `json:\"id\"`\n",
    "\tSays string `json:\"says\"`\n",
    "}\n",
    "\n",
    "type Cat struct {\n",
    "\tId              int    `json:\"id\"`\n",
    "\tSays            string `json:\"says\"`\n",
    "\tTimesMisbehaved int    `json:\"incidents\"`\n",
    "}\n",
    "\n",
    "func (_ Dog) GetEndpoint() string {\n",
    "\treturn \"/api/dogs/\"\n",
    "}\n",
    "\n",
    "func (_ Cat) GetEndpoint() string {\n",
    "\treturn \"/api/cats/\"\n",
    "}\n",
    "\n",
    "// mock function emulating an API call\n",
    "func callEndpoint(endpoint string) []byte {\n",
    "\tswitch endpoint {\n",
    "\tcase \"/api/dogs/\":\n",
    "\t\treturn []byte(`{\"says\": \"meow\", \"incidents\": 12}`)\n",
    "\tcase \"/api/cats/\":\n",
    "\t\treturn []byte(`{\"says\": \"woof\"`)\n",
    "\tdefault:\n",
    "\t\treturn []byte(`{}`)\n",
    "\t}\n",
    "}\n",
    "\n",
    "func GetAnimal[A HasEndpoint](id int) A {\n",
    "\tvar animal A\n",
    "\tfmt.Printf(\"Getting Animal ID %d at %s...\\n\", id, animal.GetEndpoint())\n",
    "\tresp := callEndpoint(animal.GetEndpoint())\n",
    "\t_ = json.Unmarshal(resp, &animal)\n",
    "\treturn animal\n",
    "}\n",
    "\n",
    "func main() {\n",
    "\tfmt.Printf(\"%v\\n\", GetAnimal[Cat](5))\n",
    "\n",
    "\tfmt.Printf(\"%v\\n\", GetAnimal[Dog](42))\n",
    "}\n",
    "```\n",
    "\n",
    "This version eliminates the runtime `switch`.\n",
    "The types are checked before runtime, and Go will yell at you if you mess something up with an unsupported type.\n",
    "`GetAnimal` is where all the magic happens.\n",
    "`A` is a type parameter that is restricted to the `HasEndpoint` interface.\n",
    "The client code is pretty simple, only requiring a type parameter to be passed to the function.\n",
    "For a thorough overview of the syntax and inner workings, please see the \"More reading\" at the end of this article.\n",
    "\n",
    "## A better Netbox client\n",
    "\n",
    "My use case for generics is a Golang Netbox client that isn't a pain in the behind to maintain.\n",
    "There are libraries based on OpenAPI or you can generate your own only to find that the schema isn't correct and complete.\n",
    "You end up adopting thousands of lines of code that needs to be maintained when you only use three or four endpoints in your client.\n",
    "\n",
    "Well, three or four endpoints grows to two dozen a few months down the road, and you are back to maintenance problem.\n",
    "I wanted a solution that would grow with the project as needed and be easy to maintain.\n",
    "As new endpoints are needed, a simple PR to add new structs is all that is needed.\n",
    "You get the client code for free!\n",
    "\n",
    "### Better abstractions\n",
    "\n",
    "Another common pattern with Netbox is that several objects have unique constraints that are not the primary key.\n",
    "When you query for an object, Netbox always returns a list.\n",
    "Many times, you simply want to queryWithFilter->GetUnique->Process.\n",
    "This is such a common use case that SQLAlchemy has a [`Query.one`](https://docs.sqlalchemy.org/en/14/orm/query.html#sqlalchemy.orm.Query.one) function that returns one item or raises an error.\n",
    "\n",
    "```go\n",
    "func (r NetboxResult[O]) First() (O, error) {\n",
    "\tvar empty O\n",
    "\tif len(r.Results) != 1 {\n",
    "\t\treturn empty, fmt.Errorf(\"expected 1 object, got %d\", len(r.Results))\n",
    "\t}\n",
    "\treturn r.Results[0], nil\n",
    "}\n",
    "```\n",
    "\n",
    "Alas, this also demonstrates a limitation of generics.\n",
    "There is no way to know what the parameterized type is at runtime without using reflection.\n",
    "In other words, it is difficult to return an error of the form \"expected 1 Rack, got 3\".\n",
    "That needs to be handled from the call site.\n",
    "This is a challenge in Python as well.\n",
    "\n",
    "## Conclusion\n",
    "\n",
    "Generics feel like a welcome addition to the Go language, and I have just scratched the surface.\n",
    "This feature does complicate your code to some degree, so it should be used sparingly.\n",
    "In my use case, I have isolated the client to its own module, and callers simply need to parameterize the function call.\n",
    "\n",
    "## More reading\n",
    "\n",
    "- [go.dev Generics Intro](https://go.dev/blog/intro-generics)\n",
    "- [go.dev Generics Tutorial](https://go.dev/doc/tutorial/generics)\n",
    "- [DO Generics Tutorial](https://www.digitalocean.com/community/tutorials/how-to-use-generics-in-go)\n"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.10"
  },
  "nikola": {
   "category": "",
   "date": "2023-03-30 06:46:23-05:00",
   "description": "",
   "link": "",
   "slug": "golang-generics",
   "tags": "go,automation",
   "title": "Golang Generics",
   "type": "text"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
